/*
    terminatorX - realtime audio scratching software
    Copyright (C) 1999  Alexander Knig
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
    File: tX_engine.c
 
    Description: Contains the code that does the real "Scratching
    		 business": XInput, DGA, Mouse and Keyboardgrabbing
		 etc. 
*/    

#include "tX_types.h"
#include "tX_engine.h"
#include <pthread.h>
#include <gtk/gtk.h>
#include <gdk/gdkprivate.h>
#include "tX_gui.h"
#include "tX_global.h"
#include "turntable.h"
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/extensions/XInput.h>
#include <X11/extensions/xf86dga.h>
#include <X11/keysym.h>
#include "tX_widget.h"

#define XWINDOW xwindow

pthread_t engine_thread=0;

pthread_mutex_t stat_lock=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t thread_lock=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t pos_lock=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t run_lock=PTHREAD_MUTEX_INITIALIZER;

int engine_status=ENG_STOPPED;

int realpos=0;

void set_real_pos(int pos)
{
	pthread_mutex_lock(&pos_lock);
	printf("setpos to %i\n", pos);
	fflush(stdout);
	realpos=pos;
	pthread_mutex_unlock(&pos_lock);
}

int get_real_pos()
{
	int pos;
	pthread_mutex_lock(&pos_lock);
	pos=realpos;
	pthread_mutex_unlock(&pos_lock);
	return(realpos);
}

int get_engine_status()
{
	int tmp;
	pthread_mutex_lock(&stat_lock);
	tmp=engine_status;
	pthread_mutex_unlock(&stat_lock);
	return(tmp);
}

void set_engine_status(int status)
{
	pthread_mutex_lock(&stat_lock);
	engine_status=status;
	pthread_mutex_unlock(&stat_lock);
}

XID OrgXPointer;
XDevice *input_device;

/* XInput grrr. strange strange....

   For now just opens and closes the selected Devices as it seems
   every device produces DGA events then.
*/

int set_xinput(Display *dpy)
{
//	XDevice *input_device=NULL;
	XDeviceInfo *devlist;			
	int listmax, i;
	int match=-1;
	
	if (globals.xinput_enable)
	{	
		devlist=XListInputDevices(dpy, &listmax);
	
		for (i=0; i<listmax; i++)
		{
			if(!strcmp(globals.xinput_device,devlist[i].name))
			{
				match=i;
			}
		
			if(devlist[i].use == IsXPointer)
			{
				OrgXPointer=devlist[i].id;
			}
		}
		
		if (match>=0)
		{
			input_device=NULL;
			input_device=XOpenDevice(dpy,devlist[match].id);
/*			if (XChangePointerDevice(dpy,input_device, 0, 1)!=Success)
			{
				match=-1;
			}*/
			XCloseDevice(dpy, input_device);
		}
		
		XFreeDeviceList(devlist);		
	
		if (match>=0) return(0); 
		else return(1);
	}
	
	return(0);
}

void reset_xinput(Display *dpy)
{
	
/*	if (globals.xinput_enable)
	{
		input_device=XOpenDevice(dpy, OrgXPointer);
		XChangePointerDevice(dpy, input_device, 0, 1);
		XCloseDevice(dpy,input_device);	
	}*/
}

void *engine(void *nil)
{
	XEvent xev;
	long mask=PointerMotionMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask;
	XMotionEvent *xmot=(XMotionEvent *) &xev;
	XKeyEvent *xkey=(XKeyEvent *) &xev;
	XButtonEvent *xbut=(XButtonEvent *) &xev;
	int quit=0;
	Time otime, ntime, dtime;
	int scratch=0;		
	Display *dpy;
	int stop_sense=globals.sense_cycles;
	KeySym key;
	f_prec warp=1.0;
			
	pthread_mutex_lock(&run_lock);
				
	dpy=XOpenDisplay(NULL); // FIXME: use correct display
	if (!dpy)
	{
		set_engine_status(ENG_ERR_XOPEN);
		pthread_exit(NULL);
	}

	if (vtt_needle_down(vttgl))
	{
		XCloseDisplay(dpy);

		set_engine_status(ENG_ERR_SOUND);
		pthread_exit(NULL);
	}
		
	if (globals.xinput_enable)
	{
		if (set_xinput(dpy))
		{
			XCloseDisplay(dpy);
			vtt_needle_up(vttgl);

			set_engine_status(ENG_ERR_XINPUT);
			pthread_exit(NULL);
		}
	}
				
//	XSelectInput(dpy, xwindow, mask);	

//	XSetInputFocus(dpy, XWINDOW, None, CurrentTime);

        if (GrabSuccess != XGrabPointer(dpy, XWINDOW, False, ButtonPressMask|ButtonReleaseMask|PointerMotionMask, GrabModeAsync,GrabModeAsync,None,None,CurrentTime))
	{
		reset_xinput(dpy);
		XCloseDisplay(dpy);
		vtt_needle_up(vttgl);

		set_engine_status(ENG_ERR_GRABMOUSE);
		pthread_exit(NULL);		
	}	
	
        if (GrabSuccess != XGrabKeyboard(dpy, XWINDOW, False, GrabModeAsync,GrabModeAsync,CurrentTime))
	{
		XUngrabPointer (dpy, CurrentTime);
		reset_xinput(dpy);		
		XCloseDisplay(dpy);

		vtt_needle_up(vttgl);
		set_engine_status(ENG_ERR_GRABKEY);
		pthread_exit(NULL);		
	}
	
	
	if (!XF86DGADirectVideo(dpy,DefaultScreen(dpy),XF86DGADirectMouse))
	{
		XUngrabKeyboard(dpy, CurrentTime);				
		XUngrabPointer (dpy, CurrentTime);
		reset_xinput(dpy);		
		XCloseDisplay(dpy);

		vtt_needle_up(vttgl);
		set_engine_status(ENG_ERR_DGA);
		pthread_exit(NULL);			
	}

	if (vttgl->mode == MODE_RECORD_SCRATCH)
	{
		vtt_reset_rec_pos(vttgl);
	}
	XAutoRepeatOff(dpy);	

	set_engine_status(ENG_RUNNING);

	otime=CurrentTime;

	while (!quit)
	{
		if (scratch)
		{
			stop_sense--;
	
			if (stop_sense<0)
			{
				vttgl->speed=0;
			}
		}
	
		if (XCheckWindowEvent(dpy, XWINDOW, mask, &xev))
		{
			switch (xev.type)
			{				
				case MotionNotify:
				if (scratch)
				{
					ntime=xmot->time;
					dtime=ntime-otime;
					if (dtime<=0) dtime=1; //these happens if you move multiple devices at the same time
					otime=ntime;
					
					if (globals.use_y)
					{
						vttgl->speed=((f_prec) xmot->y_root/dtime)*warp*globals.mouse_speed;
					}
					else
					{
						vttgl->speed=((f_prec) xmot->x_root/dtime)*warp*globals.mouse_speed;					
					}
					stop_sense=globals.sense_cycles;
				}
				break;
				
				case ButtonPress:
					switch(xbut->button)
					{
						case 1:
						vttgl->speed=0;
						scratch=1;							
						break;
						
						case 3:
						vttgl->mute_scratch=1;
						break;
						
						case 2:
						quit=1;
						break;
					}
				break;
				
				case ButtonRelease:
					switch(xbut->button)
					{
						case 1:
						vttgl->speed=globals.vtt_default_speed;
						scratch=0;							
						break;
						
						case 3:
						vttgl->mute_scratch=0;
						break;
					}
				break;
				
				case KeyPress:
					key=XKeycodeToKeysym (dpy, xkey->keycode, 0);
				
					switch(key)
					{
						case XK_space:
						vttgl->speed=0;
						scratch=1;
						break;
					
						case XK_Escape:
						case XK_Return:
						quit=1;
						break;
					
						case XK_m:
						case XK_Control_L:
						case XK_Control_R:						
						vttgl->mute_scratch=1;
						break;
						
						case XK_Alt_L:
						case XK_Alt_R:
						vttgl->mute_scratch=0;
						break;
						
						case XK_f:
						warp=10.0;						
					}
					break;
				
				case KeyRelease:
					key=XKeycodeToKeysym (dpy, xkey->keycode, 0);
				
					switch(key)
					{
						case XK_space:
						vttgl->speed=globals.vtt_default_speed;
						scratch=0;
						break;
					
						case XK_Escape:
						quit=1;
						break;
					
						case XK_m:
						case XK_Control_L:
						case XK_Control_R:						
						vttgl->mute_scratch=0;
						break;
						
						case XK_Alt_L:
						case XK_Alt_R:
						vttgl->mute_scratch=1;
						break;
						
						case XK_f:
						warp=1.0;
						break;
					}
					break;
			}
		}
		if (vtt_block_action(vttgl)) quit=2;
	}

	XF86DGADirectVideo(dpy,DefaultScreen(dpy),0);

	XUngrabKeyboard(dpy, CurrentTime);		
	XUngrabPointer (dpy, CurrentTime);
	XAutoRepeatOn(dpy);

	reset_xinput(dpy);	
	
	XCloseDisplay(dpy);

	if (vttgl->mode == MODE_RECORD_SCRATCH)
	{
		vtt_store_rec_pos(vttgl);
	}
	
	vtt_needle_up(vttgl);

	if (quit==1) set_engine_status(ENG_STOPPED);
	else set_engine_status(ENG_FINISHED);

	pthread_mutex_unlock(&run_lock);
		
	pthread_exit(NULL);
}

int run_engine()
{
	pthread_mutex_lock(&run_lock);
	
	pthread_mutex_lock(&thread_lock);

	if (engine_thread)
	{
		pthread_mutex_unlock(&thread_lock);
		return(ENG_ERR_BUSY);
	}
		
	pthread_create(&engine_thread, NULL, engine, NULL);
	
	if (!engine_thread)
	{
		pthread_mutex_unlock(&thread_lock);
		return(ENG_ERR_THREAD);
	}

	gtk_label_set(GTK_LABEL(GTK_BUTTON(action_btn)->child), "Stop");	

	pthread_detach(engine_thread);

	set_engine_status(ENG_INIT);
	
	pthread_mutex_unlock(&thread_lock);
	
	pthread_mutex_unlock(&run_lock);
		
	return (ENG_RUNNING);
}

int stop_engine()
{
	void *ret;
	
	pthread_mutex_lock(&thread_lock);	
	if (!engine_thread)
	{
		pthread_mutex_lock(&thread_lock);
		return(1);
	}
	
	pthread_join(engine_thread, &ret);
	
	engine_thread=0;
	
	pthread_mutex_unlock(&thread_lock);
	
	gtk_label_set(GTK_LABEL(GTK_BUTTON(action_btn)->child), "Start");	
	
	return (0);
}
